/*
    SunDog: xml.cpp. Simple XML engine
    Copyright (C) 2002 - 2007  Zolotov Alexandr

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
//*** Contact info: Zolotov Alexandr (NightRadio project)
//***               Ekaterinburg. Russia.
//***               Email: nightradio@gmail.com
//***               WWW: warmplace.ru

#include "../../core/core.h"
#include "../../core/debug.h"
#include "../../memory/memory.h"
#include "../../filesystem/v3nus_fs.h"
#include "../../window_manager/wmanager.h"
#include "../xml.h"

void xml_init( xml_data *x )
{
    //Create XML tree:
    x->xt_size = NODES_PER_PAGE * sizeof( xml_node );
    x->xt = (char*)mem_new( HEAP_DYNAMIC, x->xt_size, "xml tree", 0 );
    //Clean it:
    mem_set( x->xt, x->xt_size, 0 );
    //Create NULL node:
    xml_node *cur_node = xml_get_node_ptr( 0, x );
    cur_node->end_of_word = true;
}

void xml_close( xml_data *x )
{
    for( int a = 0; a < x->xt_size / (int)sizeof( xml_node ); a++ )
    {
	xml_node *cur_node = xml_get_node_ptr( a, x ); //Get current node
	if( cur_node->name ) mem_free( cur_node->name );
	if( cur_node->childs ) mem_free( cur_node->childs );
	if( cur_node->data ) mem_free( cur_node->data );
    }
    if( x->xt ) mem_free( x->xt );
}

int xml_get_value( int node, xml_data *x )
{
    xml_node *cur_node = xml_get_node_ptr( node, x );
    if( cur_node && cur_node->end_of_word == true ) return cur_node->value;
    return 0;
}

int xml_get_type( int node, xml_data *x )
{
    xml_node *cur_node = xml_get_node_ptr( node, x );
    if( cur_node && cur_node->end_of_word == true ) return cur_node->type;
    return 0;
}

void xml_set_value( int node, int value, unsigned int type, xml_data *x )
{
    xml_node *cur_node = xml_get_node_ptr( node, x );
    if( cur_node && cur_node->end_of_word == true ) 
    {
	cur_node->value = value;
	if( type ) cur_node->type = type;
    }
}

void xml_create_data( int node, int xsize, int ysize, xml_data *x )
{
    xml_node *cur_node = xml_get_node_ptr( node, x );
    if( cur_node )
    {
	if( cur_node->data ) mem_free( cur_node->data );
	cur_node->data = (char*)mem_new( HEAP_DYNAMIC, xsize * ysize + sizeof( data_header ), "data", 0 );
	data_header *h = (data_header*)cur_node->data;
	h->xsize = xsize;
	h->ysize = ysize;
    }
}

void xml_resize_data( int node, int xsize, int ysize, xml_data *x )
{
    xml_node *cur_node = xml_get_node_ptr( node, x );
    if( cur_node )
    {
	if( cur_node->data )
	{
	    cur_node->data = (char*)mem_resize( cur_node->data, xsize * ysize + sizeof( data_header ) );
	}
	data_header *h = (data_header*)cur_node->data;
	h->xsize = xsize;
	h->ysize = ysize;
    }
}

data_header *xml_get_data_header( int node, xml_data *x )
{
    xml_node *cur_node = xml_get_node_ptr( node, x );
    if( cur_node )
    {
	if( cur_node->data )
	{
	    return (data_header*)cur_node->data;
	}
	else return 0;
    }
    return 0;
}

void *xml_get_data_ptr( int node, xml_data *x )
{
    xml_node *cur_node = xml_get_node_ptr( node, x );
    if( cur_node )
    {
	if( cur_node->data )
	{
	    return cur_node->data + sizeof( data_header );
	}
	else return 0;
    }
    return 0;
}

void xml_delete_data( int node, xml_data *x )
{
    xml_node *cur_node = xml_get_node_ptr( node, x );
    if( cur_node )
    {
	if( cur_node->data ) mem_free( cur_node->data );
	cur_node->data = 0;
    }
}

xml_node *xml_get_node_ptr( int node, xml_data *x )
{
    if( node * sizeof( xml_node ) >= (unsigned)x->xt_size ) return 0;
    char *ptr = x->xt + node * sizeof( xml_node );
    return (xml_node*)ptr;
}

int xml_create_node( xml_data *x )
{
    //Search for empty nodes:
    for( int a = 0; a < x->xt_size / (int)sizeof( xml_node ); a++ )
    {
	xml_node *cur_node = xml_get_node_ptr( a, x ); //Get current node
	bool not_free = false;
	if( cur_node->end_of_word == false && cur_node->busy == false ) //It must be not "end of word" and not "busy"
	{
	    for( int b = 0; b < NUMBER_OF_CHARS; b++ ) 
		if( cur_node->links[ b ] != 0 )
		{
		    not_free = true;
		    break;
		}
	}
	else not_free = true;
	if( !not_free ) return a; //If this node is free then return it
    }
    //Can't find free node.
    //Ok. Lets add new space for it:
    int old_size = x->xt_size;
    x->xt_size += NODES_PER_PAGE * sizeof( xml_node );
    x->xt = (char*)mem_resize( x->xt, x->xt_size );
    mem_set( x->xt + old_size, x->xt_size - old_size, 0 ); //Clean new node
    return old_size / sizeof( xml_node );
}

int xml_find_object( int parent_node, char *name, xml_data *x )
{
    int p = 0;
    unsigned int c;
    int link = 0;
    int retval = 0;
    int cur_node_link = 0;
    bool created = false;

    if( name && name[ 0 ] != 0 )
    {
	c = (unsigned int)name[ p ];
	xml_node *cur_node = xml_get_node_ptr( parent_node, x );
	if( cur_node->links[ c - FIRST_CHAR ] > 0 )
	{
	    //Char exists:
	    link = cur_node->links[ c - FIRST_CHAR ];
	}
	else
	{
	    //Create this char:
	    link = xml_create_node( x );
	    created = true;
	    //Save link to it:
	    cur_node = xml_get_node_ptr( parent_node, x );
	    cur_node->links[ c - FIRST_CHAR ] = link;
	}

	//Go to the next char:
	for(;;)
	{
	    p++;
	    c = (unsigned int)name[ p ];
	    cur_node = xml_get_node_ptr( link, x );
	    cur_node_link = link;
	    if( c == 0 ) 
	    {
		//It's end of string:
		if( !cur_node->end_of_word ) created = true;
		cur_node->end_of_word = true;
		retval = link;
		break;
	    }
	    else
	    {
		if( cur_node->links[ c - FIRST_CHAR ] > 0 )
		{
		    //Char exists:
		    link = cur_node->links[ c - FIRST_CHAR ];
		}
		else
		{
		    //Create this char:
		    cur_node->busy = true;
		    link = xml_create_node( x );
		    created = true;
		    cur_node = xml_get_node_ptr( cur_node_link, x );
		    cur_node->busy = false;
		    //Save link to it:
		    cur_node->links[ c - FIRST_CHAR ] = link;
		}
	    }
	}
    }

    if( created )
    {
	//New object was created:
	//Save object name to the last node:
	xml_node *new_node = xml_get_node_ptr( retval, x );
	if( new_node->name ) mem_free( new_node->name );
	new_node->name = (char*)mem_new( HEAP_DYNAMIC, mem_strlen( name ) + 1, name, 0 );
	mem_copy( new_node->name, name, mem_strlen( name ) + 1 );

	//Also this object must add himself to the childs table
	int a;
	char pname[ 256 ];
	int parent = 0;
	for( a = mem_strlen( name ) - 1; a >= 0; a-- ) if( name[ a ] == '.' ) break; //Get last '.'
	if( a > 0 )
	{
	    //Our object has a parent:
	    int b;
	    for( b = 0; b < a; b++ ) pname[ b ] = name[ b ];
	    pname[ b ] = 0;
	    //now the pname is the parent name
	    //Get parent ID:
	    parent = xml_find_object( 0, pname, x ); //Get parent ID (and create it, if parent not exists)
	}
	//Get parent object:
	new_node = xml_get_node_ptr( parent, x );
	//Add our object to childs table:
	if( new_node->childs == 0 ) 
	{
	    new_node->childs = (int*)mem_new( HEAP_DYNAMIC, CHILDS_PER_PAGE * sizeof( int ), "childs", 0 );
	    mem_set( new_node->childs, CHILDS_PER_PAGE * sizeof( int ), 0 );
	}
	bool exists = false;
	for( a = 0; a < mem_get_size( new_node->childs ) / (int)sizeof( int ); a++ )
	{
	    if( new_node->childs[ a ] == retval ) exists = true;
	}
	if( !exists )
	{
	    for( a = 0; a < mem_get_size( new_node->childs ) / (int)sizeof( int ); a++ )
	    {
		if( new_node->childs[ a ] == 0 ) break;
	    }
	    if( a == mem_get_size( new_node->childs ) / sizeof( int ) )
	    {
		//Not enought space:
		new_node->childs = (int*)mem_resize( new_node->childs, mem_get_size( new_node->childs ) + CHILDS_PER_PAGE * 4 );
	    }
	    new_node->childs[ a ] = retval;
	    new_node->childs_num++;
	}
    }

    return retval;
}

/*

//###################################
//### FUNCTIONS FOR XML LOADING:  ###
//###################################

char *g_ts1 = 0;     //Temp string 1
char *g_ts2 = 0;     //Temp string 2
int g_p1 = 0;        //Pointer 1
int g_p2 = 0;        //Pointer 2
int g_cur_ts = 0;    //Current ts number
int g_prev_type = 0;
int g_cur_type = 0;  //Current char type

enum CHAR_TYPES
{
    C_OTHER = 0,
    C_OPENTAG,
    C_CLOSETAG,
    C_TAGEND,
    C_EQUAL,
    C_QUOTE,
    C_NAME,
    C_EOF
};

int type_of_char( int c )
{
    if( c == '<' ) return C_OPENTAG;
    else
    if( c == '>' ) return C_CLOSETAG;
    else
    if( c == '/' ) return C_TAGEND;
    else
    if( c == '=' ) return C_EQUAL;
    else
    if( c == '"' ) return C_QUOTE;
    else
    if( c >= 0x21 && c <= 0x7E ) return C_NAME;
    else
    if( c == -1 ) return C_EOF;
    else
    return 0;
}

//Load next word from a file
char *fnext_word( FILE *f )
{
    int max_size1 = mem_get_size( g_ts1 );
    int max_size2 = mem_get_size( g_ts2 );
    char *retval = 0;
    if( max_size1 > 0 && max_size2 > 0 )
    {
	for(;;)
	{
	    //Get next char
	    int c = getc( f );
	    if( c == -1 ) break;
	    int t = type_of_char( c );
	    if( t != g_prev_type )
	    {
		//We was got a char of other type:
		if( g_cur_ts == 0 )
		{
		    if( g_p1 >= max_size1 )
		    {
			//Out of temp string bound:
			g_ts1[ g_p1 - 1 ] = 0;
			return 0;
		    }
		    else g_ts1[ g_p1 ] = 0;
		    g_p2 = 0;
		    if( g_p2 >= max_size2 )
		    {
			//Out of temp string bound:
			g_ts2[ g_p2 - 1 ] = 0;
			return 0;
		    }
		    else g_ts2[ g_p2 ] = c;
		    g_p2++;
		    retval = g_ts1;
		}
		else
		{
		    if( g_p2 >= max_size2 )
		    {
			//Out of temp string bound:
			g_ts2[ g_p2 - 1 ] = 0;
			return 0;
		    }
		    else g_ts2[ g_p2 ] = 0;
		    g_p1 = 0;
		    if( g_p1 >= max_size1 )
		    {
			//Out of temp string bound:
			g_ts1[ g_p1 - 1 ] = 0;
			return 0;
		    }
		    else g_ts1[ g_p1 ] = c;
		    g_p1++;
		    retval = g_ts2;
		}
		g_cur_ts++;
		g_cur_ts &= 1;
		g_cur_type = g_prev_type;
		g_prev_type = t;
		return retval;
	    }
	    
	    //Save current char:
	    if( g_cur_ts == 0 )
	    {
		if( g_p1 >= max_size1 )
		{
		    //Out of temp string bound:
		    g_ts1[ g_p1 - 1 ] = 0;
		    return 0;
		}
		else g_ts1[ g_p1 ] = c;
		g_p1++;
	    }
	    else
	    {
		if( g_p2 >= max_size2 )
		{
		    //Out of temp string bound:
		    g_ts2[ g_p2 - 1 ] = 0;
		    return 0;
		}
		g_ts2[ g_p2 ] = c;
		g_p2++;
	    }
	}
    }
    return 0;
}

void xml_load( char *filename, xml_data *x )
{
    FILE *f = fopen( filename, "rb" );
    if( f )
    {
	g_ts1 = (char*)mem_new( HEAP_DYNAMIC, 2048, "temp string 1 for XML loading", 0 );
	g_ts2 = (char*)mem_new( HEAP_DYNAMIC, 2048, "temp string 2 for XML loading", 0 );
	g_p1 = 0;
	g_p2 = 0;
	g_prev_type = 0;
	if( g_ts1 && g_ts2 )
	{
	    for(;;)
	    {
		if( feof( f ) ) break;
		char *next_word = fnext_word( f );
		if( next_word )	prints( next_word );
	    }
	    mem_free( g_ts1 );
	    mem_free( g_ts2 );
	}
	fclose( f );
    }
}

//###################################
//### FUNCTIONS FOR XML SAVING:   ###
//###################################

//Write string to file. "^" = line break symbol
void fstr( FILE *f, char *str )
{
    int p = 0;
    for(;;)   
    {
	int c = str[ p ];
	if( c == 0 ) break;
	if( c == '^' )
	{
	    fputc( 0xD, f );
	}
	else fputc( c, f );
	p++;
    }
}

//Write value to file
void fval( FILE *f, int val )
{
    char str[ 64 ];
    int_to_string( val, str );
    fstr( f, str );
}

//Write quoted value to file
void fval_q( FILE *f, int val )
{
    char str[ 64 ];
    int_to_string( val, str );
    fstr( f, "\"" );
    fstr( f, str );
    fstr( f, "\"" );
}

//Write last child's name from string to file
void fstr_child( FILE *f, char *str )
{
    int p = 0;
    for( p = mem_strlen( str ) - 1; p >= 0; p-- )
    {
	if( str[ p ] == '.' ) break;
    }
    p++;
    for(;;)
    {
	int c = str[ p ];
	if( c == 0 ) break;
	fputc( c, f );
	p++;
    }
}

void xml_save_level( FILE *f, int obj_num, int level, xml_data *x )
{
    int a, b;
    xml_node *cur_node = xml_get_node_ptr( obj_num, x );
    bool has_data = false;
    data_header *h;

    for( b = 0; b < level; b++ ) fstr( f, "  " );
    if( cur_node->name ) 
    {
	fstr( f, "<" ); fstr_child( f, cur_node->name ); //Tag name
	if( cur_node->value ) { fstr( f, " val=" ); fval_q( f, cur_node->value ); } //Value
	//Binary Data (tag parameters):
	if( cur_node->data )
	{
	    h = (data_header*)cur_node->data;
	    has_data = true;
	    fstr( f, " xsize=" ); fval_q( f, h->xsize );
	    fstr( f, " ysize=" ); fval_q( f, h->ysize );
	    fstr( f, " type=\"hex\"" );
	}
	//Tag end:
	if( cur_node->childs || has_data ) fstr( f, ">^" ); else fstr( f, "/>^" );
	//Binary Data (content):
	if( cur_node->data )
	{
	    int ptr = 0;
	    for( a = 0; a < h->ysize; a++ )
	    {
		for( b = 0; b < h->xsize; b++ )
		{
		    uchar val = (uchar)cur_node->data[ sizeof( data_header ) + ptr ];
		    int val1 = val >> 4;
		    int val2 = val & 0xF;
		    if( val1 < 10 ) fputc( val1 + 0x30, f ); else fputc( val1 - 10 + 0x41, f );
		    if( val2 < 10 ) fputc( val2 + 0x30, f ); else fputc( val2 - 10 + 0x41, f );
		    ptr++;
		}
		fstr( f, "^" );
	    }
	}
    }
    
    if( cur_node->childs )
    {
	for( a = 0; a < mem_get_size( cur_node->childs ) / (int)sizeof( int ); a++ )
	{
	    if( cur_node->childs[ a ] )
	    {
		xml_save_level( f, cur_node->childs[ a ], level + 1, x );
	    }
	}
    }

    if( cur_node->name ) 
    {
	if( cur_node->childs || has_data )
	{
	    for( b = 0; b < level; b++ ) fstr( f, "  " );
	    fstr( f, "</" ); fstr_child( f, cur_node->name ); fstr( f, ">^" );
	}
    }
}

void xml_save( char *filename, char *description, xml_data *x )
{
    FILE *f = fopen( filename, "wb" );
    if( f )
    {
	fstr( f, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>^^" ); 
	fstr( f, "<XMLBody description=\"" );
	fstr( f, description );
	fstr( f, "\">^^" );

	xml_save_level( f, 0, 0, x );
	
	fstr( f, "^</XMLBody>^" );
	fclose( f );
    }
}

*/